/* MausConnect: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1994 1995 Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   smtp.c: Routinen fr smtp (simple mail transfer protocol)
*/

#include "mausconsys.h"

#include <stdio.h>
#include <assert.h>
#include <syslog.h>

#include "mausconlib.h"
#include "mauscon.h"
#include "metamsg.h"
#include "rfc.h"
#include "msg_io.h"

#define CRLF "\r\n"

struct smtp_private 
{
	BOOLEAN batch;
	FILE *in;
	FILE *out;
	char *host;
	char *fname;
	char smtprepbuffer[1024];
#ifdef WANT_SMTP
	long esmtp_size;
	BOOLEAN esmtp_vrfy;
	BOOLEAN esmtp_expn;
	BOOLEAN esmtp_xdsn;
	BOOLEAN esmtp_help;
	BOOLEAN esmtp_8bitmime;
#endif
};

static mio_retcode_t bsmtp_open(const char *fname, const char *mode,
	mio_t **neu, uostr_t *errbuf);
static mio_retcode_t smtp_close(mio_t *chan);
static mio_retcode_t smtp_export(mio_t *chan,metamsg_t *);
#if defined(WANT_SMTP)
static mio_retcode_t smtp_open(const char *host, const char *mode,
	mio_t **neu, uostr_t *errbuf);
static char *format_reply(struct smtp_private *);
static mio_retcode_t retcode_and_log(struct smtp_private *,int r, const char *);
static int smtp_get_reply (struct smtp_private *priv);
#endif


#if defined(WANT_SMTP)
mio_base_t mio_smtp=
{
	"smtp",

	smtp_open,
	NULL,         /* kein Import zur Zeit mglich */
	smtp_export,
	smtp_close,

	0
};
#endif

mio_base_t mio_batch_smtp=
{
	"bsmtp",

	bsmtp_open,
	NULL,
	smtp_export,
	smtp_close,

	0
};

#if defined(WANT_SMTP)
static char *
format_reply(struct smtp_private *priv)
{
	char *p=priv->smtprepbuffer;
	char *p2;
	while (isspace(*p) && isdigit(*p))
		p++;
	p2=strlen(p)+p-1;
	while (*p2=='\r' || *p2=='\n')
		*p2--='\0';
	return p;
}

static int
smtp_get_reply (struct smtp_private *priv)
{
	fd_set fds;
	struct timeval timeout;
	int cont = 1;

	FD_ZERO (&fds);
	FD_SET (fileno (priv->in), &fds);

	/* Initialize the timeout data structure. */
	/* RFC 1047 sagt, da man zumindest nach DATA *lange* warten sollte,
	 * weil die Nachrichtenverarbeitung durchaus eine Weile dauern kann.
	 * Zitat:
	 *   the increasing round-trip times on the Internet, and that some
	 *   processing after the final dot is required, the timeout for reply to
	 *   the final dot should probably be at least 5 minutes and a timeout of
	 *   10 minutes would not be unreasonable.
	 */

	timeout.tv_sec = 10 * 60;			/* seconds */
	timeout.tv_usec = 0;

	/* `select' returns 0 if timeout, 1 if input available, -1 if error. */
	if (select (FD_SETSIZE, NULL, &fds, NULL, &timeout)!=1)
	{
		*priv->smtprepbuffer='\0';
		return 421;				/* Timeout signalisieren */
	}

	while (cont)
	{
		fgets (priv->smtprepbuffer, sizeof (priv->smtprepbuffer), priv->in);
		if (priv->smtprepbuffer[3] != '-')
			cont = 0;
	}
	return (int) strtol (priv->smtprepbuffer, NULL, 10);
}

static int
smtp_get_ehlo_reply (struct smtp_private *priv)
{
	fd_set fds;
	struct timeval timeout;
	int cont;

	FD_ZERO (&fds);
	FD_SET (fileno (priv->in), &fds);

	/* Initialize the timeout data structure. */

	timeout.tv_sec = 1 * 60;			/* seconds */
	timeout.tv_usec = 0;

	/* `select' returns 0 if timeout, 1 if input available, -1 if error. */
	if (select (FD_SETSIZE, NULL, &fds, NULL, &timeout)!=1)
	{
		*priv->smtprepbuffer='\0';
		return 421;				/* Timeout signalisieren */
	}

	if (fgets (priv->smtprepbuffer, sizeof (priv->smtprepbuffer), priv->in)
		==NULL)
	{
		*priv->smtprepbuffer='\0';
		return 421;				/* Timeout signalisieren */
	}
	if (strtol(priv->smtprepbuffer,NULL,10)!=250) {
		return strtol(priv->smtprepbuffer,NULL,10);
	}
	/* auf den nchsten Zeilen stehen die untersttzten Erweiterungen */
	if (priv->smtprepbuffer[3]=='-') cont=1;
	else cont=0;
	while (cont)
	{
		char *p;
		if (fgets (priv->smtprepbuffer, sizeof (priv->smtprepbuffer), priv->in)
			==NULL) {
			*priv->smtprepbuffer='\0';
			return 421;				/* Timeout signalisieren */
		}
		p=strlen(priv->smtprepbuffer)+priv->smtprepbuffer-1;
		if (*p=='\n')
			*p--=0;
		if (*p=='\r')
			*p--=0;
		if (priv->smtprepbuffer[3] != '-') {
			p=priv->smtprepbuffer+3;
			if (!*p)
				break;
			p++;
			if (!*p)
				break;
			cont = 0;
		} else {
			p=priv->smtprepbuffer+4;
			if (!*p)
				continue;
		}
		while(isspace(*p))
			p++;
		if (0==strcasecmp(p,"8BITMIME")) {
			priv->esmtp_8bitmime=TRUE;
		} else if (0==strcasecmp(p,"EXPN")) {
			priv->esmtp_expn=TRUE;
		} else if (0==strcasecmp(p,"VRFY")) {
			priv->esmtp_vrfy=TRUE;
		} else if (0==strcasecmp(p,"XDSN")) {
			priv->esmtp_xdsn=TRUE;
		} else if (0==strcasecmp(p,"HELP")) {
			priv->esmtp_help=TRUE;
		} else if (0==strncasecmp(p,"SIZE",4)) {
			p+=4;
			while(isspace(*p))
				p++;
			if (*p)
				priv->esmtp_size=strtol(p,NULL,10);
			else
				priv->esmtp_size=0;
		}
	}
	return (int) strtol (priv->smtprepbuffer, NULL, 10);
}

#endif

static mio_retcode_t 
smtp_close(mio_t *chan)
{
	struct smtp_private *priv;
	assert(chan);
	priv=(struct smtp_private *) chan->priv_data;
	assert(priv);

#if defined(WANT_SMTP)
	if (!priv->batch)
	{
		/* echtes smtp */
		fprintf (priv->out, "QUIT" CRLF);
		(void) smtp_get_reply(priv);
		fclose(priv->out);
		fclose(priv->in);
	}
	else
#endif
	{
		/* wir schreiben in ein Batchfile ... */
		fputs ("QUIT" CRLF,priv->out);
		fclose(priv->out);
	}
	free(priv->fname);
	free(priv->host);
	free(priv);
	chan->priv_data=NULL;
	mio_free(chan);
	return MIO_OK;
}

#if defined(WANT_SMTP)
static mio_retcode_t
smtp_open(const char *host, const char *mode, mio_t **ret_handle, uostr_t *errbuf)
{
	int ret;
	char *errtext=NULL;
	struct smtp_private *priv;
	mio_t *neu;

	*ret_handle=NULL;

	if (strcmp(mode,"w"))
	{
		uostr_xdup_cstrmulti(errbuf,"illegal mode for smtp_open: ",mode,NULL);
		uostr_x0(errbuf);
		log(LOG_ERR,"%s",errbuf->data);
		return MIO_ERROR;
	}

	neu=mio_alloc(&mio_smtp);
	if (!neu)
	{
		uostr_xdup_cstr(errbuf,"out of memory");
		uostr_x0(errbuf);
		log(LOG_ERR,"%s",errbuf->data);
		return MIO_RETRY;
	}

	priv=malloc(sizeof(struct smtp_private));
	if (priv)
		priv->host=strdup(host);
	if (!priv || !priv->host)
	{
		free(priv);
		free(neu);
		uostr_xdup_cstr(errbuf,"out of memory");
		uostr_x0(errbuf);
		log(LOG_ERR,"%s",errbuf->data);
		return MIO_RETRY;
	}

	priv->in=NULL;
	priv->out=NULL;
	priv->fname=NULL;
	priv->batch=FALSE;
	priv->esmtp_size=-1;
	priv->esmtp_vrfy=0;
	priv->esmtp_expn=0;
	priv->esmtp_xdsn=0;
	priv->esmtp_help=0;
	priv->esmtp_8bitmime=0;

	if (!fopen_service ("smtp", host,&priv->in,&priv->out,&errtext))
	{
		free(neu);
		free(priv);
		free(priv->host);

		if (!errtext)
			uostr_xdup_cstr(errbuf,"out of memory");
		else 
			uostr_xdup_cstr(errbuf,errtext);
		uostr_x0(errbuf);
		log(LOG_ERR,"%s",errbuf->data);
		return MIO_RETRY;
	}

	setvbuf (priv->in, NULL, _IOLBF, 0);
	setvbuf (priv->out, NULL, _IOLBF, 0);

	ret = smtp_get_reply (priv);	/* Connection establishment reply */
	if (ret == 521)
	{
		/* RFC 1846:
		 * If an SMTP client encounters a host in an MX record that issues a 
		 * 521 greeting message, it must do one of the following two options:
		 *
		 * a) Attempt to deliver it to a different MX host for that domain.
		 * b) Return the mail with an appropriate non-delivery report.
		 *
		 * wir verwenden Variante b - MX-Lookups sind nicht implementiert.
		 */
		uostr_xdup_cstr(errbuf,"remote host doesn't accept mail");
		uostr_x0(errbuf);
		log(LOG_ERR,"%s",errbuf->data);
		fclose (priv->in);
		fclose (priv->out);
		free(priv->host);
		free(priv);
		free(neu);
		return MIO_ERROR;
	}

	if (ret != 220)
	{
		char num[STR_ULONG];
		/* Tja, keine Verbindung ... */
		str_ulong(num,ret);
		uostr_xdup_cstrmulti(errbuf,"Not welcome! returncode != 220: ",num,NULL);
		uostr_x0(errbuf);
		log(LOG_ERR,"%s",errbuf->data);
		fclose (priv->in);
		fclose (priv->out);
		free(priv->host);
		free(priv);
		free(neu);
		return MIO_RETRY;
	}

	/*
	 * Schade, sendmail mag die naechste Zeile nicht.
	 * fprintf (f, "HELO %s%s" CRLF, opts.pollsite, 
	 *    opts.mausdomain);
	 */
	fprintf (priv->out, "EHLO %s" CRLF, opts.mysitedomain);
	ret = smtp_get_ehlo_reply (priv);	/* EHLO reply */
	if (ret == 500) {
		fprintf (priv->out, "HELO %s" CRLF, opts.mysitedomain);
		ret = smtp_get_reply (priv);	/* HELO reply */
	}
	if (ret >= 300)
	{
		char num[STR_ULONG];
		str_ulong(num,ret);
		uostr_xdup_cstrmulti(errbuf,"HELO failed: ",num,NULL);
		uostr_x0(errbuf);
		log(LOG_ERR,"%s",errbuf->data);
		fclose (priv->in);
		fclose (priv->out);
		free(priv->host);
		free(priv);
		free(neu);
		return MIO_RETRY;
	}
	neu->priv_data=priv;
	*ret_handle=neu;
	return MIO_OK;
}
#endif

static mio_retcode_t
bsmtp_open(const char *fname, const char *mode, 
	mio_t **ret_handle, uostr_t *errbuf)
{
	struct smtp_private *priv;
	mio_t *neu;

	if (strcmp(mode,"w")) {
		uostr_xdup_cstrmulti(errbuf,"illegal mode for smtp_open: ",mode,NULL);
		uostr_x0(errbuf);
		log(LOG_ERR,"%s",errbuf->data);
		return MIO_ERROR;
	}

	neu=mio_alloc(&mio_batch_smtp);
	if (!neu)
	{
		uostr_xdup_cstr(errbuf,"out of memory");
		uostr_x0(errbuf);
		log(LOG_ERR,"%s",errbuf->data);
		return MIO_RETRY;
	}

	priv=malloc(sizeof(struct smtp_private));
	if (priv)
		priv->fname=strdup(fname);
	if (!priv || !priv->fname)
	{
		if (priv)
			free(priv);
		free(neu);
		uostr_xdup_cstr(errbuf,"out of memory");
		uostr_x0(errbuf);
		log(LOG_ERR,"%s",errbuf->data);
		return MIO_RETRY;
	}

	priv->in=NULL;
	priv->batch=TRUE;
	priv->host=NULL;

	priv->out=fopen(fname,"w");
	if (!priv->out)
	{
		int errno1=errno;
		uostr_xdup_cstrmulti(errbuf,"cannot open ",fname,": ",strerror(errno1),NULL);
		uostr_x0(errbuf);
		log(LOG_ERR,"%s",errbuf->data);
		free(priv);
		free(priv->fname);
		free(neu);
		return MIO_ERROR;
	}
	setvbuf (priv->out, NULL, _IOFBF, 0);

	fprintf (priv->out, "HELO %s%s" CRLF, opts.pollsite,  opts.mausdomain);

	neu->priv_data=priv;
	*ret_handle=neu;
#ifdef WANT_SMTP
	priv->esmtp_size=-1;
	priv->esmtp_vrfy=0;
	priv->esmtp_expn=0;
	priv->esmtp_xdsn=0;
	priv->esmtp_help=0;
	priv->esmtp_8bitmime=0;
#endif
	return MIO_OK;
}

#ifdef WANT_SMTP
static mio_retcode_t 
retcode_and_log(struct smtp_private *priv, int r, const char *msgid)
{
	const char *host=priv->host;

	switch(r)
	{
	case 421: /* Service not available, closing transmission channel */
		/* auch timeout! */
		log(LOG_ERR, _("connection lost"));
		return MIO_RETRY;
	case 500: /* Syntax error, command unrecognized */
	case 501: /* Syntax error in parameters or arguments */
	case 502: /* Command not implemented */
	case 503: /* Bad sequence of commands */
	case 504: /* Command parameter not implemented */
	case 550: /* user unknown */
		/* Weitermachen sinnlos. */
		log(LOG_ERR,"%s (%s)",format_reply(priv),host);
		return MIO_ERROR;
	case 521: /* host does not accept mail */
		msglog("? %s: SMTP %s (%s)",msgid, format_reply(priv),host);
		return MIO_RETRY; /* Auf einem anderen Weg nochmal versuchen */
	case 551: /* User not local, please try <forward-path> */
		msglog("? %s: SMTP %s (%s)",msgid, format_reply(priv),host);
		return MIO_RETRY; /* Auf einem anderen Weg nochmal versuchen */
	case 552: /* exceeded storage allocation */
	case 553: /* mailbox name not allowed */
	case 554: /* Transaction failed */
	case 450: /* mailbox unavailable / busy */
	case 451: /* local error in processing */
	case 453: /* insufficient system storage */
		msglog("? %s: SMTP %s (%s)",msgid, format_reply(priv),host);
		return MIO_RETRY;
	default: /* Was macht man mit unbekannten Meldungen? */
		msglog("? %s: SMTP %s (%s)",msgid, format_reply(priv),host);
		log(LOG_WARNING,_("unknown reply '%s' (%s)"),format_reply(priv),host);
		if (r>=500)
			return MIO_ERROR;
		return MIO_RETRY;
	}
}
#endif /* WANT_SMTP */

#define GET_REPLY() ((priv->batch) ? 200 : smtp_get_reply(priv))
static mio_retcode_t
smtp_export(mio_t *chan, metamsg_t *meta)
{
#ifdef WANT_SMTP
	int ret;
#endif
	FILE *out;
	FILE *in;
	struct smtp_private *priv;
	BOOLEAN size_used=FALSE;

	assert(chan);
	assert(meta);
	priv=(struct smtp_private *) chan->priv_data;

	in=priv->in;
	out=priv->out;

	if (!out)
	{
		log(LOG_ERR,_("Aieee - no smtp filehandle!"));
		mio_set_errtext(chan, _("no smtp filehandle!"), NULL);
		msglog("  %s: no SMTP filehandle (can't happen)",meta->message_id);
		return MIO_RETRY;
	}

	/* SMTP-Envelope */
	fprintf (out, "MAIL FROM:<%s>", 
		 meta->env_from ? meta->env_from : "");
	/* we do not really support 7bit - in the case the server doesn't
	 * understand 8bit, *just send 8*. Yessir, should change that.
	 */
#ifdef WANT_SMTP
	if (!priv->batch && priv->esmtp_8bitmime) {
		fprintf (out, " BODY=%s", meta->bodystat.achtbit ? "8BITMIME" : "7BIT");
	}
	/* try to use message size declarations if we read the message from
	 * a file.
	 */
	if (!priv->batch && priv->esmtp_size!=-1 && meta->file) {
		struct stat st;
		long pos=ftell(meta->file);
		if (0==fstat(fileno(meta->file),&st) && pos!=-1 && st.st_size >0) {
			long size=st.st_size - pos;
			/* etwas mehr rechnen - so sagt es auch RFC 1563:
		However, since exact computation of the message size may be
		infeasable, the client may use a heuristically-derived estimate.
		Such heuristics should be chosen so that the declared message size is
		usually larger than the actual message size. (This has the effect of
		making the counting or non-counting of SMTP DATA dots largely an
		academic point.)
			 */
			size+=(size/10);
			fprintf (out, " SIZE=%ld", size);
			size_used=TRUE;
		}
	} else if (!priv->batch && priv->esmtp_size!=-1) {
		/* estimate size */
		long size=meta->bodystat.bytes;
		size+=(size/10);
		size+=4096; /* Header - should be ok */
		fprintf (out, " SIZE=%ld", size);
		size_used=TRUE;
	}
#endif
	fputs (CRLF,out); /* terminate line */
#ifdef WANT_SMTP
	ret = GET_REPLY();	/* MAIL reply */
	if (ret > 299)
	{
		mio_set_errtext(chan, format_reply(priv), NULL);
		return retcode_and_log(priv,ret,meta->message_id);
	}
#endif
	fprintf (out, "RCPT TO:<%s>" CRLF, meta->env_to);
#ifdef WANT_SMTP
	ret = GET_REPLY();	/* MAIL reply */
	if (ret>299)
	{
		mio_set_errtext(chan, format_reply(priv), NULL);
		return retcode_and_log(priv,ret,meta->message_id);
	}
#endif

	fprintf (out, "DATA" CRLF);
#ifdef WANT_SMTP
	ret = GET_REPLY(); 
	/* Bei irgendeiner Form von Fehler die Nachricht
	 * in einen Mailbatch schreiben. Es gibt nichts, was
	 * sonst getan werden kann.
	 */
	if (ret>399)
	{
		mio_set_errtext(chan, format_reply(priv), NULL);
		return retcode_and_log(priv,ret,meta->message_id);
	}
#endif
	/* Nachricht schreiben, Zeilentrenner ist CRLF, Punkte escapen */
	if (meta->file) {
		char *lptr=NULL;
		size_t size;
		size_t got;
		while ((got=getdelim(&lptr,&size,'\n',meta->file))!=(size_t)-1) {
			lptr[got-1]=0;
			if (lptr[got-2]=='\r')
				lptr[got-2]=0;
			if (*lptr=='.')
				fputc('.',out);
			fputs(lptr,out);
			fputs(CRLF,out);
		}
		free(lptr);
	} else
		rfc_print_message(meta,out,CRLF,1);
	fprintf (out, "." CRLF);
#ifdef WANT_SMTP
	ret = GET_REPLY();
	if (ret>299)
	{
		mio_set_errtext(chan, format_reply(priv), NULL);
		return retcode_and_log(priv,ret,meta->message_id);
	}
	if (!priv->batch)
		msglog("  %s: SMTP sent (%s)",meta->message_id,priv->host);
	else
#endif
		msglog("  %s: SMTP sent (batch)",meta->message_id);
	return MIO_OK;
}
#undef GET_REPLY
