#include <unistd.h>
#include <errno.h>
#include <stdlib.h> /* getenv */
#include <string.h>
#include <signal.h>
#include <sys/wait.h>
#include "uostr.h"
#include "uoio.h"
#include "str_ulong.h"
#include "mtalib.h"

struct priv {
	int goterrs;
	uoio_t mess;
	uoio_t back;
	pid_t pid;
};
static mta_t * mta_inews_create P__((void));
static mta_ret_t mta_inews_write P__((mta_t *mta,const char *s, size_t len));
static mta_ret_t mta_inews_close P__ ((mta_t *mta));
static mta_ret_t mta_inews_abort P__ ((mta_t *mta));
static mta_ret_t mta_inews_open P__ ((mta_t **,const char *,const char *, const char *));

mta_t mta_inews={
	mta_inews_open,
	mta_inews_write,
	mta_inews_abort,
	mta_inews_close,
	0
};

static mta_t *
mta_inews_create(void)
{
	struct priv *pr;
	mta_t *m;
	pr=malloc(sizeof(struct priv));
	if (!pr) return 0;
	m=malloc(sizeof(mta_t));
	if (!m) {free(pr); return 0;}
	memcpy(m,&mta_inews,sizeof(mta_t));
	memset(pr,0,sizeof(*pr));
	m->priv=pr;
	return m;
}

#define DEFAULTINEWS "/usr/bin/inews"
static mta_ret_t
mta_inews_open(mta_t **m, const char *path,const char *from, const char *to)
{
	int pipe_back[2];
	int pipe_mess[2];
	int e;
	pid_t pid;
	mta_t *mta;
	(void) from; (void) to; /* use them */

	mta=mta_inews_create(); if (!mta) {e=errno; goto cleanup_nothing; }
	if (pipe(pipe_back) == -1) { e=errno; goto cleanup0; }
	if (pipe(pipe_mess) == -1) { e=errno;goto cleanup1; }

	pid = vfork();
	if (pid==-1) { e=errno; goto cleanup2; }
	else if (pid==0) {
		union { char *c;const char *d;} disqualify,d2,d3;
		char *inews[4];
		const char *pc=getenv("INEWSPATH");
		if (!pc) pc = path;
		if (!pc || 0==strcmp(pc,"DEFAULT")) pc=DEFAULTINEWS;

		disqualify.d=pc;
		d2.d="-h";
		d3.d="-O";
		inews[0]=disqualify.c;
		inews[1]=d2.c;
		inews[2]=d3.c;
		inews[3]=0;
		close(pipe_back[0]);
		close(pipe_mess[1]);
		if (-1==dup2(pipe_back[1],1)) _exit(120);
		if (-1==dup2(pipe_back[1],2)) _exit(120);
		if (-1==dup2(pipe_mess[0],0)) _exit(120);
		execv(disqualify.d,inews);
		_exit(120);
	}
	close(pipe_back[1]);
	close(pipe_mess[0]);
	uoio_assign_w(&((struct priv *)mta->priv)->mess,pipe_mess[1],write,0);
	uoio_assign_r(&((struct priv *)mta->priv)->back,pipe_back[0],read,0);
	((struct priv *)mta->priv)->pid=pid;
	*m=mta;
	return MTA_SUCCESS;
  cleanup2: close(pipe_mess[0]); close(pipe_mess[1]); 
  cleanup1: close(pipe_back[0]); close(pipe_back[1]); 
  cleanup0: 
    free(mta->priv);
	free(mta);
  cleanup_nothing:
	errno=e; 
	mta_errtxt=strerror(errno);
	mta_hcmssc="(#4.3.0)";
	return MTA_DEFER;
}

static mta_ret_t
mta_inews_write(mta_t *mta,const char *s, size_t len)
{
	if (-1==uoio_write_mem(&((struct priv *)mta->priv)->mess,s,len)) 
		((struct priv *)mta->priv)->goterrs++;
	if (((struct priv *)mta->priv)->goterrs) {
		mta_errtxt=strerror(errno); mta_hcmssc="(#4.3.0)";
		return MTA_DEFER;
	}
	return MTA_SUCCESS;
}

static mta_ret_t
mta_inews_abort(mta_t *mta)
{
	struct priv *priv=mta->priv;
	int status;
	kill(priv->pid,SIGTERM); 
	waitpid(priv->pid,&status,0);
	close(priv->mess.fd);
	uoio_destroy(&priv->mess);
	close(priv->back.fd);
	uoio_destroy(&priv->back);
	free(priv);
	free(mta);
	return MTA_SUCCESS;
}

static mta_ret_t
mta_inews_close(mta_t *mta)
{
	int status;
	int exitcode;
	int goterrs;
	int got_errtext=0;
	uostr_t s;
	pid_t pid;
	uoio_t i;
	struct priv *priv=mta->priv;
	s.data=0;

	if (priv->goterrs) { mta_errtxt="I/O error"; mta_hcmssc="(#4.3.0)"; }
	if (-1==uoio_flush(&priv->mess)) {priv->goterrs++;mta_errtxt=strerror(errno);mta_hcmssc="(#4.3.0)"; }
	close(priv->mess.fd);
	uoio_destroy(&priv->mess);

	while (!priv->goterrs) {
		char *t;
		ssize_t l;
		l=uoio_getdelim_zc(&priv->back,&t,'\n');
		if (l==0) break;
		if (l==-1) {mta_set_err("read:",0,"(#4.3.0)",errno); priv->goterrs++; got_errtext=1;}
		else if (l>1) {
			if ((s.data && !uostr_add_mem(&s," _/ ",4))
				|| !uostr_add_mem(&s,t,l-1)) {
				mta_set_err("out of memory",0,"(#4.3.0)",0); priv->goterrs++; got_errtext=1;
			}
		}
	}
	if (s.data && !got_errtext) {
		if (!uostr_add_mem(&s,"",1)) {
			mta_set_err("out of memory",0,"(#4.3.0)",0); priv->goterrs++; got_errtext=1;
		} else {
			mta_set_err(s.data,0,"(#4.3.0)",0); got_errtext=1;
		}
	}

	i=priv->back;
	pid=priv->pid;
	goterrs=priv->goterrs;
	free(priv);
	free(mta);

	if (waitpid(pid,&status,0)!=pid) { 
		close(i.fd); uoio_destroy(&i);
		mta_errtxt="waitpid failed.";mta_hcmssc="(#4.3.0)"; 
		return MTA_DEFER;
	}
	close(i.fd); uoio_destroy(&i);
	if (WIFSIGNALED(status)) {
		mta_errtxt="queue program crashed.";mta_hcmssc="(#4.3.0)"; return MTA_DEFER;};
	exitcode = WEXITSTATUS(status);

	switch (exitcode) {
	case 0: if (!goterrs) { 
				static char pidbuf[20]; 
				str_ulong(pidbuf,pid);
				mta_errtxt=pidbuf; /* magic: on MTA_SUCCESS the ..._close() puts an id into the errtxt */
				return MTA_SUCCESS;
			}
			return MTA_DEFER;
	default:
		if (!got_errtext) {mta_errtxt="permanent problem";mta_hcmssc="(#5.3.0)"; }
		return MTA_FATAL;
	}
}
